<?php

require_once 'system_config.php';
require_once 'gui_common.php';
require_once 'dunemdb_cache.php';
require_once 'properties.php';

///////////////////////////////////////////////////////////////////////////

class DataModel
{
    private static $instance = null;

    public static function get_instance()
    {
        if (!isset(self::$instance))
            self::$instance = new DataModel();
        return self::$instance;
    }

    ///////////////////////////////////////////////////////////////////////

    const MOVIE_INDEX_CONFIG_FILE = '/config/movie_index_config.xml';
    const STORAGE_LIST_FILE = '/tmp/run/storage_list.xml';
    const NETWORK_FOLDERS_FILE = '/config/network_folders.properties';
    const NETWORK_MOUNT_LIST_FILE = '/tmp/run/network_mount_list.xml';

    const INDEXING_STATUS_UNKNOWN = "unknown";
    const INDEXING_STATUS_NOT_STARTED = "not_started";
    const INDEXING_STATUS_RUNNING = "running";
    const INDEXING_STATUS_CANCELING = "canceling";
    const INDEXING_STATUS_CANCELED = "canceled";
    const INDEXING_STATUS_DISCONNECTED = "disconnected";
    const INDEXING_STATUS_FAILED = "failed";
    const INDEXING_STATUS_COMPLETED = "completed";

    const ALL_GENRES_GENRE_ID = "__all_genres__";
    const MOVIES_SINGLE_GENRE_ID = "__movies_single__";
    const TV_SERIES_GENRE_ID = "__tv_series__";
    const SEARCH_GENRE_ID = "__search__";
    const YEARS_GENRE_ID_PREFIX = "__years__";

    const SMB_INT_ID = 0;
    const NFS_INT_ID = 1;
    const SMB_STRING_ID = "smb";
    const NFS_STRING_ID = "nfs";

    const UDP_INT_ID = 0;
    const TCP_INT_ID = 1;
    const UDP_STRING_ID = "udp";
    const TCP_STRING_ID = "tcp";

    const STATUS_MOUNTING = "mounting";
    const STATUS_MOUNTED = "mounted";
    const STATUS_ERROR = "error";

    const FILTER_ALL_FILES = 'all_files';
    const FILTER_NOT_CONFIRMED = 'not_confirmed';
    const FILTER_CONFIRMED = 'confirmed';
    const FILTER_NOT_ASSOCIATED = 'not_associated';
    const FILTER_ASSOCIATED = 'associated';

    public static function indexing_status_is_finished($status)
    {
        return $status == self::INDEXING_STATUS_COMPLETED ||
            $status == self::INDEXING_STATUS_CANCELED ||
            $status == self::INDEXING_STATUS_FAILED ||
            $status == self::INDEXING_STATUS_DISCONNECTED;
    }

    public static function ext_status_is_error($ext_status)
    {
        return
            $ext_status == 'no_space_in_flashdata' ||
            $ext_status == 'no_space_in_persistfs';
    }

    private $folders;
    private $max_folder_id;

    private $genre_ids;
    private $genre_movie_ids;
    private $files_by_movie_id;
    private $search_token_by_movie_id;
    private $order_by_movie_id;
    private $file_by_folder_and_path;
    private $extid_by_movie_id;

    private $all_movie_ids_set;
    private $movie_path_set;

    private $stats_by_folder_id;

    private $unknown_files_by_folder_id;
    private $unknown_series_files_by_folder_id;
    private $unknown_series_max_id_by_folder_id;

    private $storages;
    private $network_folders;
    private $network_mounts;

    ///////////////////////////////////////////////////////////////////////

    private static $index_restored = false;

    public static $with_persistfs = false;

    ///////////////////////////////////////////////////////////////////////

    private static function do_determine_storage_dir()
    {
        self::$with_persistfs = false;

        $doc = file_get_contents(self::STORAGE_LIST_FILE);
        if ($doc === false)
            return;
        $xml = simplexml_load_string($doc);
        if ($xml === false)
            return;
        if (!isset($xml->storages))
            return;
        foreach ($xml->storages->children() as $storage)
        {
            $elt_name = $storage->getName();
            if ($elt_name != 'storage')
                continue;

            if (intval($storage->is_persistfs_capable) == 1 &&
                strval($storage->mount_path) == '/persistfs')
            {
                self::$with_persistfs = true;
                break;
            }
        }
    }

    public static function determine_storage_dir()
    {
        self::do_determine_storage_dir();
        hd_print("Using storage dir: ".self::get_storage_dir());
    }

    public static function get_storage_dir($ndx=0)
    {
        if ($ndx == 1)
            return self::$with_persistfs ? "/flashdata" : null;
        else
            return self::$with_persistfs ? "/persistfs" : "/flashdata";
    }

    ///////////////////////////////////////////////////////////////////////

    private function __construct()
    { 
        $dir = self::get_movie_index_dir();
        if (!is_dir($dir))
        {
            hd_print("Creating dir: $dir");
            if (mkdir($dir, 0777) === FALSE)
                hd_print("[data_model] Error creating directory '$dir'");
        }

        $dir = self::get_movie_indexer_logs_dir();
        if (!is_dir($dir))
        {
            hd_print("Creating dir: $dir");
            if (mkdir($dir, 0777) === FALSE)
                hd_print("[data_model] Error creating directory '$dir'");
        }
    }

    ///////////////////////////////////////////////////////////////////////

    public function get_progress_path($folder_id)
    {
        return"/tmp/run/movie_indexer_progress_$folder_id.txt";
    }

    public function get_progress_config_path($folder_id)
    {
        return"/config/movie_indexer_progress_$folder_id.txt";
    }

    public function get_cancel_path($folder_id)
    {
        return"/tmp/run/movie_indexer_cancel_$folder_id.txt";
    }

    ///////////////////////////////////////////////////////////////////////

    private function get_log_file_path($folder_id)
    {
        return "/tmp/run/movie_indexer_$folder_id.log";
    }

    private function get_old_log_file_path($folder_id)
    {
        return "/tmp/run/movie_indexer_$folder_id.log.old";
    }

    private function get_log_gz_file_path($folder_id, $ndx=0)
    {
        return $this->get_movie_indexer_logs_dir($ndx) .
            "/movie_indexer_$folder_id.log.gz";
    }

    private function get_old_log_gz_file_path($folder_id, $ndx=0)
    {
        return $this->get_movie_indexer_logs_dir($ndx) .
            "/movie_indexer_$folder_id.log.old.gz";
    }

    private function get_unknown_files_path($folder_id, $ndx=0)
    {
        return self::get_movie_index_dir($ndx) . "/unknown_files.$folder_id.json";
    }

    private function get_unknown_series_files_path($folder_id, $ndx=0)
    {
        return self::get_movie_index_dir($ndx) . "/unknown_series_files.$folder_id.json";
    }

    private function get_folder_movies_path($folder_id, $ndx=0)
    {
        return self::get_movie_index_dir($ndx) . "/folder_movies.$folder_id.json";
    }

    private function get_folder_genre_movies_path($folder_id, $ndx=0)
    {
        return self::get_movie_index_dir($ndx) . "/folder_genre_movies.$folder_id.json";
    }

    private function get_folder_files_path($folder_id, $ndx=0)
    {
        return self::get_movie_index_dir($ndx) . "/folder_files.$folder_id.json";
    }

    private function ensure_unknown_folder_files($folder_id)
    {
        if (isset($this->unknown_files_by_folder_id[$folder_id]))
            return;

        $this->unknown_files_by_folder_id[$folder_id] = array();

        $path = $this->get_unknown_files_path($folder_id);
        if (!is_file($path) && self::$with_persistfs)
            $path = $this->get_unknown_files_path($folder_id, 1);
        if (!is_file($path))
            return;

        $doc = file_get_contents($path);
        if ($doc === false)
        {
            hd_print("Failed to load $path.");
            return;
        }

        $json = json_decode($doc);
        if (!isset($json))
        {
            hd_print("Error parsing $path.");
            return;
        }

        if (is_array($json))
        {
            $unknown_files = &$this->unknown_files_by_folder_id[$folder_id];
            foreach ($json as $f)
            {
                if (!isset($f->path) || !isset($f->hash))
                    continue;
                $unknown_files[$f->path] = (object) array(
                    'hash' => $f->hash,
                    'ctype' => isset($f->ctype) ? $f->ctype : 0);
            }
        }
    }

    private function ensure_unknown_folder_series_files($folder_id)
    {
        if (isset($this->unknown_series_files_by_folder_id[$folder_id]))
            return;

        $this->unknown_series_files_by_folder_id[$folder_id] = array();
        $this->unknown_series_max_id_by_folder_id[$folder_id] = 0;

        $path = $this->get_unknown_series_files_path($folder_id);
        if (!is_file($path) && self::$with_persistfs)
            $path = $this->get_unknown_series_files_path($folder_id, 1);
        if (!is_file($path))
            return;

        $doc = file_get_contents($path);
        if ($doc === false)
        {
            hd_print("Failed to load $path.");
            return;
        }

        $json = json_decode($doc);
        if (!isset($json))
        {
            hd_print("Error parsing $path.");
            return;
        }

        if (!is_array($json))
            return;

        foreach ($json as $f)
        {
            $this->unknown_series_files_by_folder_id[$folder_id][$f->path] =
                (object) array(
                    's' => isset($f->s) ? $f->s : -1,
                    'e' => $f->e,
                    'ep_name' => isset($f->ep_name) ? $f->ep_name : '',
                    'series_id' => $f->series_id,
                );

            if (preg_match('/^deleted_movie:([0-9]*)$/', $f->series_id, $matches))
            {
                $n = intval($matches[1]);
                if ($n > $this->unknown_series_max_id_by_folder_id[$folder_id])
                    $this->unknown_series_max_id_by_folder_id[$folder_id] = $n;
            }
        }
    }

    private function get_unknown_folder_files($folder_id, &$folder_files)
    {
        $this->ensure_unknown_folder_files($folder_id);
        $this->ensure_unknown_folder_series_files($folder_id);

        foreach ($this->unknown_files_by_folder_id[$folder_id] as $path => $ufile)
        {
            $season = -1;
            $episode = -1;
            $ep_name = '';
            $series_id = null;
            if (isset($this->unknown_series_files_by_folder_id[$folder_id][$path]))
            {
                $f = &$this->unknown_series_files_by_folder_id[$folder_id][$path];
                $season = isset($f->s) ? $f->s : -1;
                $episode = $f->e;
                $ep_name = isset($f->ep_name) ? $f->ep_name : '';
                $series_id = $f->series_id;
            }

            $folder_files[] = (object) array(
                'path' => $path,
                'hash' => $ufile->hash,
                'ctype' => $ufile->ctype,
                's' => $season,
                'e' => $episode,
                'ep_name' => $ep_name,
                'series_id' => $series_id,
            );
        }
    }

    private function load_folder_movies_file($folder_id)
    {
        $path = $this->get_folder_movies_path($folder_id);
        if (!is_file($path) && self::$with_persistfs)
            $path = $this->get_folder_movies_path($folder_id, 1);
        if (!is_file($path))
            return;

        $doc = file_get_contents($path);
        if ($doc === false)
        {
            hd_print("Failed to load $path.");
            return;
        }

        $folder_movies = json_decode($doc);
        if (!isset($folder_movies))
        {
            hd_print("Error parsing $path.");
            return;
        }

        foreach ($folder_movies as $entry)
        {
            if (!isset($entry->id))
                continue;

            $movie_id = $entry->id;
            if (!isset($this->order_by_movie_id[$movie_id]))
            {
                $this->order_by_movie_id[$movie_id] =
                    isset($entry->order) ? $entry->order : $movie_id;
            }
            if (!isset($this->search_token_by_movie_id[$movie_id]))
            {
                $this->search_token_by_movie_id[$movie_id] =
                    isset($entry->search_token) ? $entry->search_token : '';
            }
        }
    }

    private function load_genre_movie_ids($folder_id)
    {
        $path = $this->get_folder_genre_movies_path($folder_id);
        if (!is_file($path) && self::$with_persistfs)
            $path = $this->get_folder_genre_movies_path($folder_id, 1);
        if (!is_file($path))
            return;

        $doc = file_get_contents($path);
        if ($doc === false)
        {
            hd_print("Failed to load $path.");
            return;
        }

        $folder_genre_movies = json_decode($doc);
        if (!isset($folder_genre_movies))
        {
            hd_print("Error parsing $path.");
            return;
        }

        foreach ($folder_genre_movies as $entry)
        {
            if (!isset($entry->genre_id))
                continue;

            $genre_id = $entry->genre_id;

            foreach ($entry->movies as $movie_obj)
            {
                if (is_object($movie_obj))
                {
                    $id = $movie_obj->id;
                    $order = $movie_obj->order;
                    $search_token = $movie_obj->order;

                    if (!isset($this->order_by_movie_id[$id]))
                        $this->order_by_movie_id[$id] = $order;
                    if (!isset($this->search_token_by_movie_id[$id]))
                        $this->search_token_by_movie_id[$id] = $search_token;
                }
                else
                {
                    $id = strval($movie_obj);
                    $order = isset($this->order_by_movie_id[$id]) ?
                        $this->order_by_movie_id[$id] : $id;
                    $search_token = isset($this->search_token_by_movie_id[$id]) ?
                        $this->search_token_by_movie_id[$id] : '';
                }

                if (!isset($this->files_by_movie_id[$id]))
                    continue;
                if (count($this->files_by_movie_id[$id]) == 0)
                    continue;

                if (!isset($this->genre_movie_ids[$genre_id][$id]))
                    $this->genre_movie_ids[$genre_id][$id] = $order;

                if (isset($this->all_movie_ids_set[$id]))
                    $this->all_movie_ids_set[$id] = $order;
            }
        }
    }

    public function set_stats_for_folder($folder_id, $num_files, $num_movies)
    {
        $this->stats_by_folder_id[$folder_id] = (object) array(
            'num_files' => $num_files, 'num_movies' => $num_movies);
    }

    private function load_index_for_folder($folder)
    {
        $this->set_stats_for_folder($folder->id, 0, 0);
        $this->file_by_folder_and_path[$folder->id] = array();

        $path = $this->get_folder_files_path($folder->id);
        if (!is_file($path) && self::$with_persistfs)
            $path = $this->get_folder_files_path($folder->id, 1);
        if (!is_file($path))
            return;

        hd_print("Loading folder index from $path...");

        $doc = file_get_contents($path);
        if ($doc === false)
        {
            hd_print("Failed to load $path.");
            return;
        }

        $folder_files = json_decode($doc);
        if (!isset($folder_files))
        {
            hd_print("Error parsing $path.");
            return;
        }

        $num_folder_files = count($folder_files);
        hd_print("... $num_folder_files files");
        if ($num_folder_files == 0)
            return;

        $file_by_path = &$this->file_by_folder_and_path[$folder->id];
        $folder_movie_ids_set = array();

        foreach ($folder_files as $folder_file)
        {
            $rel_path = $folder_file->path;
            if (isset($file_by_path[$rel_path]))
                continue;

            $hash = $folder_file->hash;
            $ctype = isset($folder_file->ctype) ? $folder_file->ctype : 0;
            $dunemdb_id = $folder_file->id;
            $season = isset($folder_file->s) ? $folder_file->s : -1;
            $episode = isset($folder_file->e) ? $folder_file->e : -1;
            $match_confirmed = isset($folder_file->match_confirmed) ?
                $folder_file->match_confirmed == "yes" : false;
            $ass_type = isset($folder_file->ass_type) ?
                $folder_file->ass_type : 0;
            $ep_name = isset($folder_file->ep_name) ? $folder_file->ep_name : '';

            if (!isset($this->files_by_movie_id[$dunemdb_id]))
                $this->files_by_movie_id[$dunemdb_id] = array();

            $this->all_movie_ids_set[$dunemdb_id] = 1;
            $folder_movie_ids_set[$dunemdb_id] = 1;

            $this->extid_by_movie_id[$dunemdb_id] = $folder_file->kzid;

            $f = (object) array
                (
                 'path' => $rel_path,
                 'hash' => $hash,
                 'ctype' => $ctype,
                 'folder_id' => $folder->id,
                 'match_confirmed' => $match_confirmed,
                 's' => $season,
                 'e' => $episode,
                 'ass_type' => $ass_type,
                 'ep_name' => $ep_name,
                 'movie_id' => $dunemdb_id,
                );
            $file_by_path[$rel_path] = $f;

            $this->movie_path_set["$f->folder_id/$f->path"] = 1;
            $this->files_by_movie_id[$dunemdb_id][] = $f;
        }

        $this->set_stats_for_folder($folder->id,
            count($folder_files), count($folder_movie_ids_set));
    }

    /////////////////////////////////////////////////////////////////////// 

    public static function cmp_ordered_movie_ids($f1, $f2)
    {
        return strnatcasecmp($f1, $f2);
    }

    public static function cmp_files($f1, $f2)
    {
        if ($f1->s != $f2->s)
            return $f1->s - $f2->s;
        if ($f1->e != $f2->e)
            return $f1->e - $f2->e;
        if ($f1->ass_type != $f2->ass_type)
            return $f1->ass_type - $f2->ass_type;
        return PathHelper::cmp_paths($f1->path, $f2->path);
    }

    private static function sort_files(&$files)
    {
        usort($files, 'DataModel::cmp_files');
    }

    public static function cmp_files_by_name($f1, $f2)
    {
        $name1 = HD::path_basename($f1->path);
        $name2 = HD::path_basename($f2->path);
        $n = strnatcasecmp($name1, $name2);
        if ($n != 0)
            return $n;

        $n = strnatcasecmp($f1->path, $f2->path);
        if ($n != 0)
            return $n;

        return $f1->folder_id - $f2->folder_id;
    }

    public static function sort_files_by_name(&$files)
    {
        usort($files, 'DataModel::cmp_files_by_name');
    }

    public static function cmp_genre_ids($a, $b)
    {
        $dunemdb_cache = DunemdbCache::get_instance();

        $other1 = $dunemdb_cache->is_other_genre_id($a);
        $other2 = $dunemdb_cache->is_other_genre_id($b);
        if ($other1 != $other2)
            return $other1 ? 1 : -1;

        return ($dunemdb_cache->get_genre_position_by_id($a) >=
            $dunemdb_cache->get_genre_position_by_id($b)) ? 1 : -1;
    }

    private function load_index()
    {
        $this->genre_movie_ids = array();
        $this->files_by_movie_id = array();
        $this->search_token_by_movie_id = array();
        $this->order_by_movie_id = array();
        $this->file_by_folder_and_path = array();
        $this->all_movie_ids_set = array();
        $this->movie_path_set = array();
        $this->stats_by_folder_id = array();
        $this->unknown_files_by_folder_id = array();
        $this->unknown_series_files_by_folder_id = array();
        $this->unknown_series_max_id_by_folder_id = array();
        $this->extid_by_movie_id = array();

        foreach ($this->folders as $folder)
            $this->load_index_for_folder($folder);

        foreach ($this->files_by_movie_id as $dunemdb_id => &$files)
            self::sort_files($files);

        foreach ($this->folders as $folder)
        {
            $this->load_folder_movies_file($folder->id);
            $this->load_genre_movie_ids($folder->id);
        }

        DunemdbCache::get_instance()->ensure_genres_loaded();
        $this->genre_ids = array_keys($this->genre_movie_ids);
        usort($this->genre_ids, 'DataModel::cmp_genre_ids');

        uasort($this->all_movie_ids_set, "DataModel::cmp_ordered_movie_ids");
        foreach ($this->genre_movie_ids as $genre_id => &$movie_ids)
            uasort($movie_ids, "DataModel::cmp_ordered_movie_ids");

        hd_print('Loaded: ' .
            count($this->genre_movie_ids) . ' genres, ' .
            count($this->all_movie_ids_set) . ' movies, ' .
            count($this->movie_path_set) . ' movie files.');

        unset($this->movie_path_set);
    }

    private function do_reload_network_folders()
    {
        $this->network_folders = array();

        hd_print('Loading network folders.');

        $path = self::NETWORK_FOLDERS_FILE;

        $p = new Properties($path, false);

        for ($n = 0;; $n++)
        {
            $key = "id.$n";
            if (!isset($p->{$key}))
                break;
            $id = trim($p->{$key});
            if ($id == '')
                continue;

            $key = "name.$n";
            if (!isset($p->{$key}))
                continue;
            $name = trim($p->{$key});
            if ($name == '')
                continue;

            $key = "type.$n";
            $type = isset($p->{$key}) ? intval($p->{$key}) : self::SMB_INT_ID;
            if ($type == self::SMB_INT_ID)
                $type = self::SMB_STRING_ID;
            else if ($type == self::NFS_INT_ID)
                $type = self::NFS_STRING_ID;
            else
                continue;

            $key = "server.$n";
            $server = trim($p->{$key});
            if ($server == '')
                continue;

            $key = "directory.$n";
            $directory = trim($p->{$key});
            if ($directory == '')
                continue;

            $key = "user.$n";
            $user = isset($p->{$key}) ? trim($p->{$key}) : "";

            $key = "password.$n";
            $password = isset($p->{$key}) ? trim($p->{$key}) : "";

            $key = "protocol.$n";
            $protocol = isset($p->{$key}) ? intval($p->{$key}) : self::UDP_INT_ID;
            if ($protocol == self::UDP_INT_ID)
                $protocol = self::UDP_STRING_ID;
            else if ($protocol == self::TCP_INT_ID)
                $protocol = self::UDP_STRING_ID;
            else
                continue;

            if ($type == self::SMB_STRING_ID)
            {
                $server = strtolower($server);
                $directory = strtolower($directory);
            }

            $this->network_folders[] = (object) array(
                'id' => $id,
                'name' => $name,
                'type' => $type,
                'server' => $server,
                'dir' => $directory,
                'user' => $user,
                'password' => $password,
                'proto' => $protocol,
            );
        }

        hd_print("The ".count($this->network_folders)." network folders loaded");
    }

    public function reload_network_mounts()
    {
        $this->network_mounts = array();

        hd_print('Loading network mount list.');

        $path = self::NETWORK_MOUNT_LIST_FILE;

        $doc = file_get_contents($path);
        if ($doc === false)
        {
            hd_print("Failed to load $path.");
            return;
        }

        $xml = simplexml_load_string($doc);
        if ($xml === false)
        {
            hd_print("Error parsing $path.");
            return;
        }

        foreach ($xml->children() as $elt)
        {
            $elt_name = $elt->getName();
            if ($elt_name != 'mount')
                continue;

            $id = intval($elt["id"]);
            $path = strval($elt["path"]);
            $type = strval($elt["type"]);
            if ($type != self::SMB_STRING_ID && $type != self::NFS_STRING_ID)
                continue;
            $server = strval($elt["server"]);
            $dir = strval($elt["dir"]);
            $user_name = isset($elt["user_name"]) ?
                strval($elt["user_name"]) : "";
            $password = isset($elt["password"]) ?
                strval($elt["password"]) : "";
            $proto = isset($elt["proto"]) ?
                strval($elt["proto"]) : "";
            $status = strval($elt["status"]);
            if ($status != self::STATUS_MOUNTED && $status != self::STATUS_MOUNTING)
                continue;
            $mount_tm = isset($elt["mount_tm"]) ?
                intval($elt["mount_tm"]) : 0;

            hd_print("Added mount: $id $path ($type $server:$dir)");

            if ($type == self::SMB_STRING_ID)
            {
                $server = strtolower($server);
                $dir = strtolower($dir);
            }

            $this->network_mounts[$id] = (object) array(
                'id' => $id,
                'path' => $path,
                'type' => $type,
                'server' => $server,
                'dir' => $dir,
                'user' => $user_name,
                'password' => $password,
                'proto' => $proto,
                'status' => $status,
                'mount_tm' => $mount_tm);
        }

        hd_print("The ".count($this->network_mounts)." mounts loaded.");
    }

    public function reload_network_folders()
    {
        // TODO
        $this->do_reload_network_folders();
    }

    private function do_reload_storage_list($silent=false)
    {
        $this->storages = array();
        if (!$silent)
            hd_print('Loading storage list.');

        $path = self::STORAGE_LIST_FILE;

        $doc = file_get_contents($path);
        if ($doc === false)
        {
            hd_print("Failed to load $path.");
            return;
        }

        $xml = simplexml_load_string($doc);
        if ($xml === false)
        {
            hd_print("Error parsing $path.");
            return;
        }

        if (!isset($xml->storages)) {
            hd_print("Warning: root element not 'storages' in $path.");
            return;
        }

        foreach ($xml->storages->children() as $storage)
        {
            $elt_name = $storage->getName();
            if ($elt_name != 'storage')
                continue;

            if (intval($storage->is_persistfs_capable) == 1)
                continue;
            if (!isset($storage->name))
                continue;

            $id = strval($storage->id);
            $kind = strval($storage->kind);
            $name = strval($storage->name);
            $uuid = strval($storage->uuid);
            $mount_path = strval($storage->mount_path);

            $arr = array(
                'id' => $id,
                'kind' => $kind,
                'name' => $name,
                'uuid' => $uuid,
                'mount_path' => $mount_path
            );

            if (isset($storage->label))
                $arr['label'] = strval($storage->label);
            if (isset($storage->user_name))
                $arr['user_name'] = strval($storage->user_name);
            if (isset($storage->user_caption))
                $arr['user_caption'] = strval($storage->user_caption);

            if (!$silent)
            {
                hd_print("Added storage: $id (kind: $kind, ".
                    "name: $name, uuid: $uuid, mount_path: $mount_path)");
            }

            $this->storages[] = (object) $arr;
        }
    }

    public function reload_storage_list($silent=false)
    {
        // TODO: skip parsing if data not changed.

        $old_storages = $this->storages;

        $this->do_reload_storage_list($silent);
    }

    public function reload_folders_configuration()
    {
        $this->folders = array();
        $this->max_folder_id = 0;

        hd_print('Loading movie folders configuration.');

        $miconfig_file = null;
        if (is_file(self::MOVIE_INDEX_CONFIG_FILE))
            $miconfig_file = self::MOVIE_INDEX_CONFIG_FILE;

        if ($miconfig_file == null)
            return;

        $doc = file_get_contents($miconfig_file);
        if ($doc === false)
        {
            hd_print("Failed to load $miconfig_file.");
            return;
        }

        $xml = simplexml_load_string($doc);
        if ($xml === false)
        {
            hd_print("Error parsing $miconfig_file.");
            return;
        }

        if (!isset($xml->max_folder_id))
        {
            hd_print("Warning: no 'max_folder_id' in $miconfig_file.");
            return;
        }
        $this->max_folder_id = intval($xml->max_folder_id['value']);
        if ($this->max_folder_id <= 0)
        {
            hd_print("Warning: invalid max_folder_id=$max_folder_id ".
                "in $miconfig_file.");
            return;
        }

        if (!isset($xml->folders)) {
            hd_print("Warning: no 'folders' in $miconfig_file.");
            return;
        }

        foreach ($xml->folders->children() as $folder)
        {
            $elt_name = $folder->getName();
            if ($elt_name != 'folder') {
                hd_print("Warning: ".
                    "unexpected element '$elt_name' inside 'folders'");
                continue;
            }

            $id = strval($folder['id']);
            if (!preg_match('/^[0-9]+$/', $id))
            {
                hd_print("Invalid folder: ".
                    "required 'id' is invalid or missing: '$id'.");
                continue;
            }

            $source_id = strval($folder['source_id']);
            if (!preg_match('/\S/', $source_id))
            {
                hd_print("Invalid folder: ".
                    "required 'source_id' is empty or missing.");
                continue;
            }

            $source_caption = isset($folder['source_caption']) ?
                strval($folder['source_caption']) : '';

            $path = strval($folder['path']);
            if (!preg_match('/\S/', $path))
            {
                hd_print("Invalid folder: ".
                    "required 'path' is empty or missing.");
                continue;
            }

            $caption = strval($folder['caption']);
            if (!preg_match('/\S/', $caption))
            {
                hd_print("Invalid folder: ".
                    "required 'caption' is empty or missing.");
                continue;
            }

            $indexing_status = isset($folder['indexing_status']) ?
                strval($folder["indexing_status"]) :
                self::INDEXING_STATUS_NOT_STARTED;

            if ($indexing_status != self::INDEXING_STATUS_RUNNING &&
                $indexing_status != self::INDEXING_STATUS_NOT_STARTED &&
                $indexing_status != self::INDEXING_STATUS_CANCELING &&
                $indexing_status != self::INDEXING_STATUS_CANCELED &&
                $indexing_status != self::INDEXING_STATUS_DISCONNECTED &&
                $indexing_status != self::INDEXING_STATUS_FAILED &&
                $indexing_status != self::INDEXING_STATUS_COMPLETED)
            {
                hd_print("Warning: ".
                    "wrong folder indexing status: '$indexing_status'");
                $indexing_status = self::INDEXING_STATUS_NOT_STARTED;
            }

            $indexing_ext_status = isset($folder['indexing_ext_status']) ?
                strval($folder['indexing_ext_status']) : '';

            $indexing_start_tm = intval($folder['indexing_start_tm']);
            if ($indexing_start_tm < 0)
                $indexing_start_tm = 0;

            $indexing_end_tm = 0;
            $indexing_finished =
                $indexing_status == self::INDEXING_STATUS_CANCELED ||
                $indexing_status == self::INDEXING_STATUS_DISCONNECTED ||
                $indexing_status == self::INDEXING_STATUS_FAILED ||
                $indexing_status == self::INDEXING_STATUS_COMPLETED;
            if ($indexing_finished)
            {
                $indexing_end_tm = intval($folder['indexing_end_tm']);
                if ($indexing_end_tm < 0)
                    $indexing_end_tm = 0;
            }

            hd_print("Added folder: #$id $caption (source: $source_id, ".
                "path: $path), indexing $indexing_status ($indexing_ext_status)");

            $this->folders[$id] = (object) array(
                'id' => $id,
                'source_id' => $source_id,
                'source_caption' => $source_caption,
                'path' => $path,
                'caption' => $caption,
                'indexing_status' => $indexing_status,
                'indexing_ext_status' => $indexing_ext_status,
                'indexing_start_tm' => $indexing_start_tm,
                'indexing_end_tm' => $indexing_end_tm,
            );
        }
    }

    public function unset_model()
    {
        $this->storages = null;
        $this->network_folders = null;
        $this->network_mounts = null;

        $this->folders = null;
        $this->max_folder_id = 0;

        $this->genre_name_by_id = null;
        $this->files_by_movie_id = null;
        $this->file_by_folder_and_path = null;
        $this->extid_by_movie_id = null;
        $this->all_movie_ids_set = null;
        $this->movie_path_set = null;
        $this->stats_by_folder_id = null;
        $this->unknown_files_by_folder_id = null;
        $this->unknown_series_files_by_folder_id = null;
        $this->unknown_series_max_id_by_folder_id = array();
    }

    public function ensure_loaded()
    {
        SystemConfig::get_instance()->ensure_loaded();

        if (isset($this->folders))
            return;

        $this->do_reload_storage_list();
        $this->do_reload_network_folders();
        $this->reload_network_mounts();

        $this->reload_folders_configuration();
        $this->load_index();

        $mem = sprintf("%.1fMb", memory_get_usage() / 1000000.0);
        $mem2 = sprintf("%.1fMb", memory_get_usage(true) / 1000000.0);
        hd_print("Used memory: $mem ($mem2 real)");
    }

    ///////////////////////////////////////////////////////////////////////

    public function get_indexer_pid_path($folder_id)
    {
        return "/tmp/run/movie_indexer_$folder_id.pid";
    }

    public function get_indexer_pid($folder_id)
    {
        $pid_path = $this->get_indexer_pid_path($folder_id);
        if (!is_file($pid_path))
            return -1;

        $str = file_get_contents($pid_path);
        if ($str === false)
            return -1;

        $pid = intval($str);
        return $pid <= 1 ? -1 : $pid;
    }

    public function indexer_is_running($folder_id)
    {
        $pid = $this->get_indexer_pid($folder_id);
        if ($pid == -1)
            return false;

        return posix_kill($pid, 0);
    }

    public static function get_same_folder_episode_files($file, &$files)
    {
        if ($file->e < 0)
            return array($file);

        $result_files = array();
        foreach ($files as $f)
        {
            if ($f->folder_id != $file->folder_id || $f->e < 0)
                continue;

            $n = PathUtils::get_common_path_prefix_len($f->path, $file->path);
            if (false !== strpos($f->path, '/', $n))
                continue;
            if (false !== strpos($file->path, '/', $n))
                continue;

            $result_files[] = $f;
        }
        return $result_files;
    }

    public static function get_same_folder_unknown_series_files(
        $path, $file, &$unknown_series_files)
    {
        if (!isset($file->series_id) || $file->e < 0)
            return array($path => $file);

        $result_files = array();
        foreach ($unknown_series_files as $p => $f)
        {
            if (!isset($f->series_id) ||
                $f->e < 0 ||
                $f->series_id != $file->series_id)
            {
                continue;
            }

            $n = PathUtils::get_common_path_prefix_len($p, $path);
            if (false !== strpos($p, '/', $n))
                continue;
            if (false !== strpos($path, '/', $n))
                continue;

            $result_files[$p] = $f;
        }
        return $result_files;
    }

    ///////////////////////////////////////////////////////////////////////

    public function &get_folders()
    {
        return $this->folders;
    }

    public function get_folder_ids()
    {
        $ids = array();
        foreach ($this->folders as $folder)
            $ids[] = $folder->id;
        return $ids;
    }

    public function get_all_genre_ids()
    {
        return $this->genre_ids;
    }

    public function get_all_genres_count()
    {
        return count($this->genre_ids);
    }

    public function get_movies_count()
    {
        return count($this->all_movie_ids_set);
    }

    public function get_genre_movie_ids($genre_id)
    {
        if ($genre_id == self::ALL_GENRES_GENRE_ID)
            return $this->all_movie_ids_set;
        return isset($this->genre_movie_ids[$genre_id]) ?
            $this->genre_movie_ids[$genre_id] : null;
    }

    public function is_collection_movie($movie_id)
    {
        return isset($this->all_movie_ids_set[$movie_id]);
    }

    public function get_files_by_movie_id($movie_id)
    {
        return isset($this->files_by_movie_id[$movie_id]) ?
            $this->files_by_movie_id[$movie_id] : array();
    }

    public function get_files_count_by_movie_id($movie_id)
    {
        return isset($this->files_by_movie_id[$movie_id]) ?
            count($this->files_by_movie_id[$movie_id]) : 0;
    }

    public static function cmp_by_path($f1, $f2)
    {
        return PathHelper::cmp_paths($f1->path, $f2->path);
    }

    public function get_all_folder_files($filter, $folder_id)
    {
        $folder_files = array();

        if ($filter != self::FILTER_NOT_ASSOCIATED)
        {
            $confirmed_only = $filter == self::FILTER_CONFIRMED;
            $not_confirmed_only = $filter == self::FILTER_NOT_CONFIRMED;

            foreach ($this->file_by_folder_and_path[$folder_id] as $file)
            {
                if ($confirmed_only && !$file->match_confirmed)
                    continue;
                if ($not_confirmed_only && $file->match_confirmed)
                    continue;

                $folder_files[] = $file;
            }
        }

        if ($filter != self::FILTER_ASSOCIATED &&
            $filter != self::FILTER_CONFIRMED)
        {
            $this->get_unknown_folder_files($folder_id, $folder_files);
        }

        usort($folder_files, 'DataModel::cmp_by_path');

        return $folder_files;
    }

    public function get_folder_files_count_ext($folder_id,
        &$unmatched, &$confirmed, &$not_confirmed)
    {
        $this->ensure_unknown_folder_files($folder_id);
        $unmatched = count($this->unknown_files_by_folder_id[$folder_id]);

        $confirmed = 0;
        $not_confirmed = 0;

        foreach ($this->file_by_folder_and_path[$folder_id] as $file)
        {
            if ($file->folder_id != $folder_id)
                continue;

            if ($file->match_confirmed)
                $confirmed++;
            else
                $not_confirmed++;
        }
    }

    public function &get_folder_by_id($folder_id)
    {
        /*
        return isset($this->folders[$folder_id]) ?
            $this->folders[$folder_id] : null;
            */

        return $this->folders[$folder_id];
    }

    public function get_folder_by_caption($caption)
    {
        foreach ($this->folders as $folder)
        {
            if ($folder->caption == $caption)
                return $folder;
        }
        return null;
    }

    public function folder_exists_with_path($source_id, $path)
    {
        foreach ($this->folders as $folder)
        {
            if ($folder->source_id == $source_id && $folder->path == $path)
                return true;
        }
        return false;
    }

    public function get_folder_by_path($source_id, $path)
    {
        foreach ($this->folders as $folder)
        {
            if ($folder->source_id == $source_id && $folder->path == $path)
                return $folder;
        }
        return null;
    }

    public function get_folder_movies_count($folder_id)
    {
        return isset($this->stats_by_folder_id[$folder_id]->num_movies) ?
            $this->stats_by_folder_id[$folder_id]->num_movies : 0;
    }

    public function get_folder_files_count($folder_id)
    {
        return isset($this->stats_by_folder_id[$folder_id]->num_files) ?
            $this->stats_by_folder_id[$folder_id]->num_files : 0;
    }

    public function get_movie_file_by_path($folder_id, $path)
    {
        $file_by_path = &$this->file_by_folder_and_path[$folder_id];
        return isset($file_by_path[$path]) ? $file_by_path[$path] : null;
    }

    public function &get_storages()
    {
        return $this->storages;
    }

    public static function get_movie_index_dir($ndx=0)
    {
        return $ndx == 1 && !self::$with_persistfs ? null :
            self::get_storage_dir($ndx) . "/movie_index";
    }

    public function get_movie_indexer_logs_dir($ndx=0)
    {
        return $ndx == 1 && !self::$with_persistfs ? null :
            self::get_storage_dir($ndx) . "/movie_indexer_logs";
    }

    public function get_storages_count()
    {
        return count($this->storages);
    }

    public function get_storage_by_name($name)
    {
        foreach ($this->storages as $storage)
        {
            if ($storage->name == $name)
                return $storage;
        }
        return null;
    }

    public function get_storage_by_id($id)
    {
        foreach ($this->storages as $storage)
        {
            if ($storage->id == $id)
                return $storage;
        }
        return null;
    }

    public function get_storage_by_source_id($source_id)
    {
        if (0 === strpos($source_id, "storage_name://"))
        {
            $storage_name = substr($source_id, 15);
            return $this->get_storage_by_name($storage_name);
        }
        return null;
    }

    public function &get_network_folders()
    {
        return $this->network_folders;
    }

    public function get_network_folders_count()
    {
        return count($this->network_folders);
    }

    public function get_network_folder_by_id($id)
    {
        foreach ($this->network_folders as $nf)
        {
            if ($nf->id == $id)
                return $nf;
        }
        return null;
    }

    private function get_network_folder_by_id_or_name($is_name, $id_or_name)
    {
        foreach ($this->network_folders as $nf)
        {
            if ($id_or_name == ($is_name ? $nf->name : $nf->id))
                return $nf;
        }
        return null;
    }

    public function get_network_folder_by_source_id($source_id)
    {
        $is_name = 0 === strpos($source_id, "network_folder://");
        $is_id = !$is_name &&
            0 === strpos($source_id, "network_folder_id://");
        if (!$is_name && !$is_id)
            return null;
        $id_or_name = substr($source_id, $is_name ? 17 : 20);
        return $this->get_network_folder_by_id_or_name($is_name, $id_or_name);
    }

    public function &get_network_mounts()
    {
        return $this->network_mounts;
    }

    public function get_network_mount_by_id($id)
    {
        foreach ($this->network_mounts as $mnt)
        {
            if ($mnt->id == $id)
                return $mnt;
        }
        return null;
    }

    public function get_network_mount_by_source($type, $server, $dir)
    {
        foreach ($this->network_mounts as $mnt)
        {
            if ($mnt->type == $type &&
                $mnt->dir == $dir &&
                ($mnt->server == $server ||
                 (isset($mnt->server_ipaddr) &&
                  $mnt->server_ipaddr == $server)))
            {
                return $mnt;
            }
        }
        return null;
    }
}

///////////////////////////////////////////////////////////////////////////
?>
